************************************************************************
      program dmrcc
************************************************************************
*     Date last modified: 14/11/2014                                   *
*     Author:      Mihaly Kallay                                       *
*     Version:     1026                                                *
*     Description: Driver for mrcc                                     *
*     Routines called: getmemr8,getmem,gconfig,intinp,initialt,mpdenom,*
*                      ccit,gcoupl                                     *
*     References: J. Chem. Phys. 115, 2945 (2001); 117, 980 (2002);    *
*                 119, 2991 (2003).                                    *
************************************************************************
#include "MRCCCOMMON"
      integer i,j,verbosity
      real*8 etest,temp,cput,walt
      character*4 ctol,cscr4
      character*8 cscr8,gopt,freq,lccrest,orblocguess,oniom !HB
      character*15 cscr15
      character*16 devnul
      character*24 basopt
      character*30 test
      logical llg,ltmp
      common/output/ devnul

c {{{ for MPI
      integer, parameter :: IMPI = 1, OMPI = 2
      character*4 mpitasks_ch
      integer mpi_rank, mpitasks
      data mpi_rank /0/
#if defined(MPI)
      integer mpi_err, nlen, icomm

      integer system, error

      integer, allocatable :: err_codes(:)
      character*8 rank_ch
      character*(MPI_MAX_PROCESSOR_NAME) pname
      character*64 mpi_argv(2)

      integer ev_status
      character(len = 32) env_var

#ifdef INTELMPI
      call get_environment_variable("I_MPI_MPIRUN", 
     $      value = env_var, status = ev_status)
      if(ev_status .eq. 1) call exec_mpirun(IMPI)
#endif
#ifdef OPENMPI
      call get_environment_variable("OMPI_COMM_WORLD_RANK", 
     $      value = env_var, status = ev_status)
      if(ev_status .eq. 1) call exec_mpirun(OMPI)
#endif

      call MPI_Init(mpi_err)
      call MPI_Comm_rank(MPI_COMM_WORLD, mpi_rank, mpi_err)

c      master_thread = .true. ! 1 dmrcc

      write(rank_ch, '(i8)') mpi_rank
      if(mpi_rank .ne. 0) then
         outfilename = 'mrcc.out.' // trim(adjustl(rank_ch))

         iout = 960
         call delete_file(outfilename, .true.)
         open(iout, file = outfilename)
      else
         iout = 6
         outfilename = '                '
         output_folder = 'mpi_output'
#ifdef WINTEL
         call execute_command_line('mkdir ' // trim(output_folder)
     $, exitstat=error)
#else
         error = system('mkdir -p ' // trim(output_folder) // char(0))
#endif
c         if(error .ne. 0) then
c            write(*, '(" Error: Couldn''t create output directory")')
c            call dmrccend(1)
c         end if
      end if
       
      if(mpi_rank .ne. 0) devnul = '>>' // trim(outfilename)

      call MPI_Get_processor_name(pname, nlen, mpi_err)

      call getcwd(wdir)

#else
      open(minpfile, file = 'MINP')
      call getkeym('mpitasks', 8, mpitasks_ch, 4)
      if(mpitasks_ch .ne. '    ') then
         read(mpitasks_ch, '(i4)') mpitasks
      else
         mpitasks = 1
      end if
#ifdef INTELMPI
      if(mpitasks .ne. 1) call exec_mpirun(IMPI)
#endif
#ifdef OPENMPI
      if(mpitasks .ne. 1) call exec_mpirun(OMPI)
#endif
      close(minpfile)

      iout = 6
      devnul = '                '
      outfilename = '                '
      call getcwd(wdir)
#endif

c }}}
C Title
      write(iout,1003)
      write(iout,"(26x,a19)") 'MRCC program system'
      write(iout,1003)
      write(iout,*) 
      write(iout,*) '                             Written by'
      write(iout,*) 'Mihaly Kallay, Peter R. Nagy, David Mester, '
     $// 'Zoltan Rolik, Gyula Samu,'
      write(iout,*) '  Jozsef Csontos, Jozsef Csoka, P. Bernat Szabo, '
     $// 'Laszlo Gyevi-Nagy,'
      write(iout,*) '   Bence Hegely, Istvan Ladjanszki, '
     $// 'Lorant Szegedy, Bence Ladoczki,'
      write(iout,*) '      Klara Petrov, Mate Farkas, Pal D. Mezei, '
     $// 'and Adam Ganyecz'
      write(iout,*) 
      write(iout,*) 
     $'       Department of Physical Chemistry and Materials Science'
      write(iout,*) 
     $'           Budapest University of Technology and Economics'
      write(iout,*) 
     $'                 Budapest P.O.Box 91, H-1521 Hungary'
      write(iout,*) 
      write(iout,*) '                             www.mrcc.hu'
      write(iout,*) 
      write(iout,*) '                   Release date: February 22, 2020'
1003  format(1x,70('*'))
#if defined(MPI)
      write(iout, *)
      write(iout, '(//," dmrcc running on ",a)') trim(pname)
#endif
C Create empty pids file (deleting it only is probably sufficient)
      open(scrfile1,file='pids')
      close(scrfile1,status='delete')
      open(scrfile1,file='pids')
      close(scrfile1)
C Remove unnecessary files
      open(scrfile1,file='ROUTE')
      close(scrfile1,status='delete')
      open(scrfile1,file='RESTART')
      close(scrfile1,status='delete')
      open(scrfile1,file='ONIOMROUTE')
      close(scrfile1,status='delete')
      call signalinit ! initialise response to SIG's of the OS
C Check if input files exist
      inquire(file='SCHINFO',exist=llg) !HB
       if(llg) call ishell("rm -f SCHINFO") !HB
      inquire(file='ITER',exist=llg)
       if(llg) call ishell("rm -f ITER")
      inquire(file='MOCOEF.CAN',exist=llg)
       if(llg) call ishell("rm -f MOCOEF.CAN")
      inquire(file='MINP',exist=llg)
      devnul='                '
       if(llg) then
        call runit('minp',cput,walt)
        open(minpfile, file = 'MINP')

#if defined(MPI)
         call getkey('mpitasks', 8, mpitasks_ch, 4)
         read(mpitasks_ch, '(i4)') mpitasks
#endif

        call getkey('lccrest',7,lccrest,8) 
        call getkey('orblocguess',11,orblocguess,8)
        inquire(file='MOCOEF.LOC',exist=ltmp)
        if(ltmp.and.
     $    .not.(lccrest.eq.'domain  '.or.lccrest.eq.'restart '.or.
     $          orblocguess.eq.'read    '.or.orblocguess.eq.'restart '))
     $             call ishell("rm -f MOCOEF.LOC") ! MOCOEF.LOC is needed for lccrest and orblocguess
        call getkey('basopt',6,basopt,24)
        call getkey('gopt',4,gopt,8)
        call getkey('freq',4,freq,8)
        call getkey('oniom',5,oniom,8) !HB
         if(basopt.ne.'off                     ') then
C Optimization of basis sets
          call getkey('verbosity',9,cscr4,4)
          read(cscr4,"(i4)") verbosity
#ifdef WINTEL
           if(verbosity.le.2) devnul='>NULL           '
#else
           if(verbosity.le.2) devnul='>/dev/null      '
#endif
          write(iout,*)
          call basoptdrv(gbasfile)
         else if(gopt.ne.'off     ') then
C Geometry optimization
          call geomoptdrv(minpfile)
         else
C Single-point calculation
!HB
          if(oniom.ne.'off') then
           call composite(llg)
          else 
           call spoint(llg)
          endif
         endif
C Harmonic vibrational frequency calculation
         if(freq.ne.'off     ') call freqdrv
       else
C Single-point calculation with the old machinery
!HB
        if(oniom.ne.'off') then
         call composite(llg)
        else
         call spoint(llg)
        endif
       endif
C Check test energy
       if(llg) then
        call getkey('test',4,test,30)
         if(test.ne.'off                           ') then
          read(test,*) etest
          call getenergy(temp,cscr15)
           if(trim(adjustl(cscr15)).eq.'SCF') then
            call getkey('scftol',6,ctol,4)
           else
            call getkey('cctol' ,5,ctol,4)
           endif
          read(ctol,*) i
           if(gopt.ne.'off     '.or.freq.ne.'off     ') i=i-1
          write(iout,*)
c          if(dabs(etest-temp).le.10.d0**(-i+1)) then
           if(dabs(etest-temp).le.10.d0**(-i)) then
            write(iout,*) 'Test job completed successfully.'
           else
            write(iout,*) 'Test job failed to complete successfully!'
            write(iout,*)
            write(iout,"(' Test energy:       ',f22.12)") etest
            write(iout,"(' Calculated energy: ',f22.12)") temp
            write(iout,"(' Difference:        ',f22.12)") etest-temp
            call dmrccend(1)
           endif
         endif
       endif
C
      call dmrccend(0)
      end
C
************************************************************************
      subroutine xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,tomem,tomemo)
************************************************************************
* Driver for xmrcc                                                     *
************************************************************************
#include "MRCCCOMMON"
      integer icyc,nactoo,nactvo,convero
      real*8 almem,almemo,remem,tomem,tomemo,cput,walt
      logical lopt,llow
C
      call runit('goldstone',cput,walt)
C
      open(scrfile6,status='unknown',file='fort.24',form='unformatted')
      rewind(scrfile6)
      write(scrfile6) icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,nacto,nactv,nocc,nvirt,conver,pert,tomem,tomemo
      close(scrfile6)
C
      call runit('xmrcc',cput,walt)
C
      open(scrfile6,status='unknown',file='fort.24',form='unformatted')
      rewind(scrfile6)
      read(scrfile6) icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,nacto,nactv,nocc,nvirt,conver,pert,tomem,tomemo
      close(scrfile6)
C
      return
      end
C
************************************************************************
      subroutine updatei(log,noccup,nbmax,nmax,lr)
************************************************************************
* Update input file                                                    *
************************************************************************
#include "MRCCCOMMON"
      integer i,j,jj,nline,iscr1(18),iscr2(2),noccup(*),nbmax,nmax
      integer occv(nbmax),actv(nbmax),exv(nmax)
      character line(1000)
      real*8 eps
      logical log,lr
C
      open(inpfile,status='unknown',file='fort.56')
      rewind(inpfile)
      read(inpfile,*) iscr1,ptfreq,iscr2!,eps
      nline=1
      iscr1(1)=op1
      iscr1(2)=nsing
      iscr1(3)=ntrip
      iscr1(4)=rest
       if(lr) then
         if(op1.ne.op) then
          iscr1(1)=op1+1
          iscr1(5)=3
         else
          iscr1(5)=1
         endif
       endif
      iscr1(7)=conver
      iscr1(10)=iclsh
      iscr1(12)=ihf
      iscr1(13)=ndoub
      iscr1(14)=nacto
      iscr1(15)=nactv
       if(nactv*nacto.ne.0.and.ssmrcc.eq.0) iscr1(17)=iscr1(1)
      read(inpfile,'(1000a1)',end=1001) line
      nline=2
      read(inpfile,*,end=1001) (occv(i),i=1,nbasis/2)
      nline=3
      read(inpfile,*,end=1001) (actv(i),i=1,nbasis/2)
      nline=4
      read(inpfile,*,end=1001) (exv(i),i=1,op)
      nline=5
1001  continue
C
      rewind(inpfile)
      write(inpfile,"(18i8,f9.5,i7,i9,i7,1pe10.3)") iscr1,ptfreq,iscr2!,eps
      jj=1
       do j=1,1000
         if(line(j).ne.' ') jj=j
       enddo
      write(inpfile,"(1000a1)") (line(j),j=1,jj)
       if(log) then
        write(inpfile,"(10000i3)") (noccup(j),j=1,nbasis/2)
       else
         if(nline.ge.3)write(inpfile,"(10000i3)") (occv(i),i=1,nbasis/2)
       endif
       if(nline.ge.4) write(inpfile,"(10000i3)") (actv(i),i=1,nbasis/2)
       if(nline.ge.5) write(inpfile,"(10000i3)") (exv(i),i=1,op)
      close(inpfile)
C
      return
      end
C
************************************************************************
      subroutine runit(string,cput,walt)
************************************************************************
* Call external programs                                               *
************************************************************************
#include "MRCCCOMMON"

#if defined(MPI)
      character(len = 4) :: localcc
      character(len = 8) :: scfalg
      integer :: mpi_rank, mpi_err
      logical lll
#endif
      integer ierr
      real*8 cput,walt
      character*(*) string
      character*16 devnul
      common/output/ devnul
C
      flush(iout)

#if defined(MPI)
      call MPI_Comm_rank(MPI_COMM_WORLD, mpi_rank, mpi_err)

      if(mpi_rank .ne. 0) close(iout)
#endif

#ifdef WINTEL
      write (iout,*) " "
      call execute_command_line(
     $'echo ************************ %DATE% %TIME%' //
     $' *************************'//trim(devnul))
      call ishell('echo  Executing ' // string//'...'//trim(devnul))
      write (iout,*) " "
#else
      call ishell('echo'//devnul)
      call ishell(
     $'echo " ************************ "`date +"%F  %T"`"' //
     $' *************************"'//trim(devnul))
      call ishell('echo " Executing ' // string//'..."'//trim(devnul))
      call ishell('echo'//devnul)
#endif

      call flush(iout)
#if defined (G95)
       if(string.eq.'goldstone') then
        call ishell('export G95_IGNORE_ENDFILE=t; exec goldstone' 
     $// trim(devnul))
       else
        call ishell('export G95_IGNORE_ENDFILE=f; '// "exec "// string 
     $// trim(devnul))
       endif
#else

#if defined(MPI)
c      call MPI_Comm_size(MPI_COMM_WORLD, mpi_size, mpi_err)
      call MPI_Barrier(MPI_COMM_WORLD, mpi_err)

      localcc='off '
      if (string.eq.'drpa') call getkey('localcc',7,localcc,4) ! cannot call getkey for string=minp

      scfalg='        '
       if(string.ne.'minp') then
        inquire(file='KEYWD',exist=lll)
         if(lll) call getkey('scfalg',6,scfalg,8)
       endif

      if(string .eq. 'ccsd' .or. 
     $   string .eq. 'mrcc' .or.
     $  (string .eq. 'scf'.and.scfalg.ne.'disk    ')) then
c     $  (string .eq. 'drpa' .and. localcc .eq. '2019')) then
         call run_mpi_program(string)
      else if(string.eq.'xmrcc') then
#ifdef WINTEL
         call ishell( "xmrcc_mpi" // trim(devnul))
#else
         call ishell( "exec xmrcc_mpi" // trim(devnul))
#endif
      else
#endif
#ifdef WINTEL
         call ishell( string // trim(devnul))
#else
         call ishell( "exec " // string // trim(devnul))
#endif
#if defined(MPI)
      end if

      call MPI_Barrier(MPI_COMM_WORLD, mpi_err)

      if(mpi_rank .ne. 0) 
     $      open(unit = iout, file = outfilename, access = 'append')
#endif

#endif
      open(unit=exitfile,file='EXIT',status='unknown')
      read(exitfile,*) ierr,cput,walt
      close(exitfile)
       if(ierr.ne.0) then
        write(iout,*)
        write(iout,*) 'Fatal error in ',string,'.'
        write(iout,*) 'Program will stop.'
        call flush(iout)
        call dmrccend(1)
       endif
C
      return
      end
C
#if defined(MPI)
************************************************************************
      subroutine run_mpi_program(string)
************************************************************************
#include "MRCCCOMMON"
      character(len = *) :: string

      logical :: flag
      integer :: request, time
      integer :: mpitasks, mpi_rank, mpi_err
c      integer, allocatable :: error_codes(:)
c      integer :: error_codes(1024)
      character(len = 4) :: mpitasks_ch

      character(len = 8) :: date
      character(len = 10) :: times
      character(len = 5) :: zone
      integer :: dt(8)
      character(len = MPI_MAX_PROCESSOR_NAME) :: pname
      integer :: nlen
      character(len = 64) :: id
      integer :: children_comm

      call MPI_Comm_rank(MPI_COMM_WORLD, mpi_rank, mpi_err)

      call getkey('mpitasks', 8, mpitasks_ch, 4)
      read(mpitasks_ch, '(i4)') mpitasks

      call MPI_Get_processor_name(pname, nlen, mpi_err)
      call date_and_time(date, times, zone, dt)

      ! create file for identification
      open(idfile, file = "dmrcc.id")
      id =  date // times // zone // trim(pname)
      write(idfile, *) trim (id); flush(idfile)
      close(idfile)

      ! spawn mpi processes
c      allocate(error_codes(mpi_size))
c      allocate(error_codes(mpitasks))
      call MPI_Comm_spawn(trim(string) // '_mpi', MPI_ARGV_NULL, 
     $      mpitasks,MPI_INFO_NULL, 0, MPI_COMM_WORLD, children_comm,
c     $      error_codes, mpi_err)
     $      MPI_ERRCODES_IGNORE, mpi_err)
c      deallocate(error_codes)

      ! communicate working directory
      if(mpi_rank .eq. 0) then
         call MPI_Bcast(wdir, 512, MPI_CHARACTER, MPI_ROOT, 
     $         children_comm, mpi_err)
      else
         call MPI_Bcast(wdir, 512, MPI_CHARACTER, MPI_PROC_NULL, 
     $         children_comm, mpi_err)
      end if

      ! communicate output directory
      if(mpi_rank .eq. 0) then
         call MPI_Bcast(output_folder, 32, MPI_CHARACTER, MPI_ROOT, 
     $         children_comm, mpi_err)
      else
         call MPI_Bcast(output_folder, 32, MPI_CHARACTER, MPI_PROC_NULL,
     $         children_comm, mpi_err)
      end if

      ! communicate id
      if(mpi_rank .eq. 0) then
         call MPI_Bcast(pname, MPI_MAX_PROCESSOR_NAME, MPI_CHARACTER, 
     $         MPI_ROOT, children_comm, mpi_err)
      else
         call MPI_Bcast(pname, MPI_MAX_PROCESSOR_NAME, MPI_CHARACTER, 
     $         MPI_PROC_NULL, children_comm, mpi_err)
      end if

      ! wait for children to finish
      call MPI_Ibarrier(children_comm, request, mpi_err)
      flag = .false.
      do while(.not. flag)
         call sleep_f(1)   ! required for OpenMPI to yield processor
         call MPI_Test(request, flag, MPI_STATUS_IGNORE, mpi_err)
      end do

      call MPI_Comm_disconnect(children_comm, mpi_err)
c      call sleep_f(2)

      call delete_file('dmrcc.id', .true.)

      end subroutine
#endif

************************************************************************
      subroutine memopt(icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,llv,nact1,noc1,nacto1,tomem,tomemo)
************************************************************************
* Optimize memory requirement                                          *
************************************************************************
#include "MRCCCOMMON"
      integer i,icyc,nactoo,nactvo,convero,nn,nact1,noc1,nacto1
      real*8 almem,almemo,remem,tomem,tomemo
      logical lopt,llow,log,llv
C
      llv=.true.
       if(almem.gt.remem.and.lopt.and.nact1.eq.0) then
        nact1=nint(dble(noc1)/2.d0)
         if(mod(nact1,2).eq.1) then
          nact1=nact1+1
           if(nact1.ge.noc1) nact1=max(0,nact1-2)
         endif
         if(nact1.gt.0) then
          call updatei(.false.,i,nbasis/2,op,.false.)
          call xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,tomem,tomemo)
         endif
       endif
C
       if(almem.gt.remem.and.lopt.and.nact1.ne.0) then
        log=.true.
        llow=.true.
        nn=nact1
         do while(llow.and.nact1.ge.4)
          nact1=nact1-2
          call updatei(.false.,i,nbasis/2,op,.false.)
          call xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,tomem,tomemo)
           if(llow) log=.false.
           if(almem.le.remem) then
            llv=.false.
            return
           endif
         enddo
         if(log) then
          llow=.true.
          nact1=nn
           do while(llow.and.nact1.le.noc1-4)
            nact1=nact1+2
            call updatei(.false.,i,nbasis/2,op,.false.)
            call xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,
     $lopt,llow,tomem,tomemo)
             if(almem.le.remem) then
              llv=.false.
              return
             endif
           enddo
         endif
        nact1=nacto1
       endif
C
      return
      end
C
************************************************************************
      subroutine ddmrcc(ccprog,csapprox)
************************************************************************
* Driver for one mrcc calling
************************************************************************
#include "MRCCCOMMON"
      integer icyc,nactoo,nactvo,convero,mm,nn,i,j,k,l,m,ilow,restold
      real*8 remem,almem,almemo,tomem,tomemo,cput,walt
      logical lopt,llow,log,llv,log1,csapprox
      character*4 ccprog
      character*5 scftype
      logical llg
C Initialize interface file
      open(unit=ifcfile,file='iface',status='replace')
      write(ifcfile,"(68a)")
     $'#property method        sym st mul        value' //
     $'             CPU(sec)       Wall(sec)'
      close(ifcfile)
C Run Lori's ccsd
      m=0
c       ccprog='    ' 
      inquire(file='MINP',exist=llg)
       if(llg) then
c         call getkey('ccprog',6,ccprog,4)
        call getkey('scftype',7,scftype,5)
         if(ccprog.eq.'ccsd') then
           if(scftype.eq.'uhf  '.or.scftype.eq.'rohf ') then
            if (csapprox) then
             call runit('ccsd',cput,walt)
            else 
            call runit('uccsd',cput,walt)
            endif
           else
            call runit('ccsd',cput,walt)
           endif
          return
         else if(ccprog.eq.'cis ') then
          call runit('cis',cput,walt)
          return
         endif
       else
        open(inp,status='unknown',file='fort.55')
        rewind(inp)
        read(inp,*,end=1005) m
1005    close(inp)
       endif
C
      lopt=.false.
      icyc=0
      almemo=2.d0**32.d0
      tomemo=2.d0**32.d0
      nactoo=0
      nactvo=0
      convero=0
      almem=0.d0
      remem=0.d0
      tomem=0.d0
      llow=.false.
      restold=rest
C Check if input files have been changed
       if(ssmrcc.gt.0.or.locno.ge.3) goto 1000
      inquire(file='fort.14',exist=log1)
       if(.not.log1) goto 1000
      inquire(file='fort.10',exist=log1)
       if(.not.log1) goto 1000
      rewind(gfile)
      read(gfile,*,end=1000) i,j,k,maxmem,l
      close(gfile)
      call ishell('cat fort.56 > fort.11')
      call ishell("sed '/E/,$d' fort.55 >> fort.11")
      call ishell(
     $"sed '1,/inpfile/d' fort.14 | diff fort.11 - | wc -l > fort.99")
      call ishell(
     $"sed '1,/inpfile/d' fort.10 | diff fort.11 - | wc -l >> fort.99")
      rewind(99)
      read(99,*,end=1000) nn
      read(99,*,end=1000) mm
      close(99,status='delete')
       if(m.gt.0.and.mm+nn.eq.0.and.j.eq.nvirt.and.k.eq.nocc) goto 1004
1000  continue
C Calculate memory requirement using the input data
      call xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,lopt,
     $llow,tomem,tomemo)
C Approximate methods
       if(almem.gt.remem.and.pert.gt.0.and.conver.ne.2.and.locno.ne.3) 
     $then
        conver=2
        call updatei(.false.,i,nbasis/2,op,.false.)
        call xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,lopt,
     $llow,tomem,tomemo)
         if(.not.llow) conver=0
       endif
C Optimizing the number of active particles
      call memopt(icyc,nactoo,nactvo,convero,almem,almemo,remem,lopt,
     $llow,llv,nactv,nvirt,nactvo,tomem,tomemo)
C Optimizing the number of active holes
      if(llv.and.locno.ne.3) call memopt(icyc,nactoo,nactvo,convero,
     $almem,almemo,remem,lopt,llow,llv,nacto,nocc,nactoo,tomem,tomemo)
C Restore the inputs for the optimal case
       if(lopt.and.(nactoo.ne.nacto.or.nactv.ne.nactvo.or.
     $convero.ne.conver.or..not.llow)) then
        nacto=nactoo
        nactv=nactvo
        conver=convero
        call updatei(.false.,i,nbasis/2,op,.false.)
        call xmrccm(icyc,nactoo,nactvo,convero,almem,almemo,remem,lopt,
     $llow,tomem,tomemo)
       endif
C
       if(almem.gt.remem) then
        write(iout,*) 
     $'Unable to optimize memory usage: Insufficient memory!'
        call flush(iout)
        call dmrccend(1)
       else if(lopt.and.abs(rest).eq.1.and.nacto+nactv.gt.0) then
        write(iout,*) 'The number of active orbitals has been changed!'
        write(iout,*) 'It is dangerous to restart the program!'
        call flush(iout)
       endif
C Save input data
#ifdef WINTEL
      call ishell('echo inpfile >>fort.14')
#else
      call ishell('echo "inpfile" >> fort.14')
#endif
      call ishell('cat fort.56 >> fort.14')
      call ishell("sed '/E/,$d' fort.55 >> fort.14")
#ifdef WINTEL
      call ishell('echo inpfile >>fort.10')
#else
      call ishell('echo "inpfile" >> fort.10')
#endif
      call ishell('cat fort.56 >> fort.10')
      call ishell("sed '/E/,$d' fort.55 >> fort.10")
C Execute lower-level calculations if rest=2 or rest=4
       if(abs(rest).eq.2.or.abs(rest).eq.4) then
        ilow=2
         if(calc.ge.2.and.calc.le.9) ilow=3
         if(nroot.gt.1.and.abs(rest).ne.4) ilow=1
         do op1=ilow,op
           if(op1.eq.ilow) then
             if(abs(rest).eq.4) then
              rest=isign(3,rest)
             else
              rest=0
             endif
           else
            rest=isign(2,rest)
           endif
          call updatei(.false.,i,nbasis/2,op,calc.eq.1.and.nroot.eq.1)
          call runit('goldstone',cput,walt)
          call runit('xmrcc',cput,walt)
           if(op1.ne.op) call runit('mrcc',cput,walt)
         enddo
       endif
C Execute mrcc
1004  continue
C #if defined (MPI) || (defined (PGF) && defined (OMP))
#if defined (MPI)
       if(ssmrcc.gt.0) then
        call runit('mrcc',cput,walt)   !!!szemet!!!
       else
c        write(iout,*)
c        write(iout,*) 'Now launch mrcc in parallel mode!'
c        write(iout,*)
         call run_mpi_program('mrcc')
       endif
#else
      call runit('mrcc',cput,walt)
#endif
C Restore input file
       if(rest.ne.restold) then
        rest=restold
        op1=op
        call updatei(.false.,i,nbasis/2,op,.false.)
       endif
C
      return
      end
C
************************************************************************
      subroutine dssmrcc(nbmax,ccprog)
************************************************************************
* Driver for SSMRCC calculations
************************************************************************
#include "MRCCCOMMON"
      integer i,nref,iref,nbmax,noccup(nbmax),roccup(nbmax)
      integer nsingo,ntripo,ndoubo,iclsho,ihfo
      real*8 coef
      logical conv,log
      character*4 ccprog
C
      call ishell('cp fort.56 bup.56')
      ss1route=.false.
      ss2route=.false.
      ss3route=.false.
C Solve FCI in reference space
      ss1route=.true.
      open(ssfile,status='unknown',file='fort.31',form='formatted')
      rewind(ssfile)
      write(ssfile,*) ss1route,ss2route,ss3route,1,1.d0
      close(ssfile)
      call ddmrcc(ccprog,.false.)
      ss1route=.false.
      rewind(ssfile)
      read(ssfile,*) nref
      call ishell('cp fort.16 fort.17')
C Generate denominators for rho's
      ss2route=.true.
c ...
      ss2route=.false.
C Iterations
      ss3route=.true.
      conv=.true.
      rewind(scfile)
       do while(conv)
         do iref=1,nref
          read(scfile,*) coef,noccup
           if(iref.eq.1) then
             do i=1,nbmax
               roccup(i)=noccup(i)
             enddo
           endif
          nsingo=nsing
          ntripo=ntrip
          ndoubo=ndoub
          iclsho=iclsh
          ihfo=ihf
           do i=1,nbmax
             if(iabs(noccup(i)).eq.1) log=.true.
           enddo
           if(log.and.ndoub.eq.0) then
            nsing=0
            ntrip=0
            ndoub=nsingo+ntripo
            iclsh=0
            ihf=0
           endif
          call updatei(.true.,noccup,nbasis/2,op,.false.)
          rewind(ssfile)
          write(ssfile,*) ss1route,ss2route,ss3route,iref,coef
          write(ssfile,*) roccup
          close(ssfile)
          call ddmrcc(ccprog,.false.)
          nsing=nsingo
          ntrip=ntripo
          ndoub=ndoubo
          iclsh=iclsho
          ihf=ihfo
         enddo
        conv=.false.
       enddo
      ss3route=.false.
      call ishell('mv bup.56 fort.56')
C
      return
      end
C
************************************************************************
      subroutine locdrv(intcpu,intwal,orbcpu,orbwal,localcc,ovirtcpud,
     $ovirtwald,lmp2ened,scfcpu,scfwal,calctype)  
************************************************************************
* Driver for local CC calculations
************************************************************************
#include "MRCCCOMMON"
      integer nbasisold,noccold,iocc,i,idegen,nocorr,nbcorr,n1,n2,n3
      integer nom,nbm,nocorrmin,nocorrmax,nbcorrmin,nbcorrmax,maxmet
      integer imet,nmet,ndom,idom,ncmo,cmo,iccsd,idomst,nbasiss,noccs 
      integer idumy,ncoi,nact,nmetact,ccmem(8),ipart,imp2,nob,noccolda
      integer noccoldb,ncorep
      parameter(maxmet=30)
      real*8 ecor,totcpu(maxmet),totwal(maxmet),ecl(maxmet),escf
      real*8 mp2corr,ovirtcpu,ovirtwal,lmp2ene,scfcpu,scfwal
      real*8 ovirtcpud,ovirtwald,lmp2ened,intcpu,intwal,intcpud,intwald
      real*8 times(6),edenergy(5)
      real*8 cccpu,ccwal,orbcpu,orbwal,mulcpu,mulwal,eenv,pairpllno
      real*8 tcpu,twal,mp2cpu,mp2wal,mp2fact
      character*3 lmp2dens,half
      character*5 scftype
      character*4 ccprog,localcc,accprog,eccprog,setccprog,talg
      character*6 core,cscr,ccorr,totcorr
      character*128 label,meth(maxmet),beg,end,ccmo,methact(maxmet)
      logical log1,lseqc,mp2env,hfenv,lrc,ccenv,lanycc,localcc15p
      logical laptoldefault,cycle_this_domain,csapprox,mp2fact_halved
      logical localccXp
      character*16 naf,envcalc,cscr16,calctype,uenvcalc,alaptol,elaptol
      character*8 naftyp,lccrest,corembed,ccsdalg,lccoporder,scfalg
C Type of correction
      ccorr='MP2   '
      half='   '
      mp2fact=1.0d0 ! MP2 correction scale for all but CCSD, for that mp2fact=0.5
      mp2fact_halved=.false.
C Execute mulli
       if(localcc.eq.'2013') then 
         call runit('mulli',mulcpu,mulwal)
         call ishell('rm -f 55.*')
       elseif (localcc15p(localcc)) then
         mulcpu=0.d0
         mulwal=0.d0
       endif  
C      
      call getkey('core',4,core,6)
      call getkey('ccprog',6,ccprog,4)
      call getkey('lmp2dens',8,lmp2dens,3)
      call getkey('scfalg',6,scfalg,8)
      call getvar('lseqc     ',lseqc)
      call getkey('lccrest',7,lccrest,8) 
      call getvar('eref      ',escf)
      call getkey('naf_cor',3,naf,16) 
      call getkey('naftyp',6,naftyp,8) 
      call getkey('talg',4,talg,4) 
      call getkey('corembed',8,corembed,8)
      call getkey('ccsdalg',7,ccsdalg,8)
      call getkey('lccoporder',10,lccoporder,8)
      call getvar('nal       ',nal) !for oslcc
      call getvar('nbe       ',nbe) !for oslcc
      call getkey('scftype',7,scftype,5) !for oslcc
      ncorep=ncore
      if (trim(core).eq.'corr') ncorep=0
      nmetact=1 ! number of invoked methods (for the active region if corembed)
      if (.not.(corembed.ne.'off     ')) envcalc='none            ' 
c {{{ read/initialize localcc.restart
      if (lccrest.eq.'on      '.or.lccoporder.eq.'lccfirst') then
        open(unit=scrfile2,file='localcc.restart',status='old')
        call localccrestart('read',scrfile2,calctype,meth,nmet,
     $   nbasis,nocc,ncore,idomst,nmet,
     $   nocorr,nbcorr,nocorrmin,nocorrmax,nbcorrmin,nbcorrmax,nact,
     $   ecl,totcpu,
     $   totwal,lmp2ene,mp2corr,pairpllno,eenv,
     $ edenergy(1),edenergy(2),edenergy(3),edenergy(4),edenergy(5),ccmem
     $   ,intcpu,intwal,scfcpu,scfwal,orbcpu,orbwal,ovirtcpu,ovirtwal)
        if (lccoporder.eq.'lccfirst') then ! do not overwrite timings with uninitilaized values of localcc.restart
           ovirtcpu=ovirtcpud ; ovirtwal=ovirtwald
        endif
        idomst=idomst+1
        write(iout,*)
        write(iout,122)
        write(iout,122)
        write(iout,*)
      else 
        idomst=1 
        edenergy(1:5)=0.d0
c read corrections from the ED calculations 
        if(localcc15p(localcc)) then
          open(unit=ifcfile,file='iface',status='old')
          read(ifcfile,*) lmp2ene,mp2corr,eenv
          close(ifcfile)
          pairpllno=mp2corr
          if (naftyp.eq.'jpqpilot'.or.naftyp.eq.'jpq     ') then 
             mp2corr=lmp2ene-eenv ! second term of mp2corr is read from the result file of ccsd for all idom values
          endif
        else
          lmp2ene=lmp2ened
          mp2corr=0.d0
          eenv=0.d0
          pairpllno=mp2corr
        endif
c inititalize localcc.restart for restarting with the first LIS 
        open(unit=scrfile2,file='localcc.restart')
        rewind(scrfile2)
        nmet=min(maxmet,30)
        ecl(1:nmet)=0.d0
        totcpu(1:nmet)=0.d0
        totwal(1:nmet)=0.d0
        ovirtcpu=ovirtcpud
        ovirtwal=ovirtwald
        ccmem(1:8)=0
        call localccrestart('init',scrfile2,calctype,meth,nmet,
     $   nbasis,nocc,ncore,0,nmet,
     $   0,0,nocc,0,nbasis,0,nocc,
     $   ecl,totcpu,
     $   totwal,lmp2ene,mp2corr,pairpllno,eenv,
     $   0.d0,0.d0,0.d0,0.d0,0.d0,ccmem,
     $   intcpu,intwal,scfcpu,scfwal,orbcpu,orbwal,ovirtcpu,ovirtwal )
      endif
c }}}
      nbasiss=nbasis 
      noccs=nocc 
      nbasisold=nbasis/2
      noccold=nocc/2
      noccolda=nal-ncorep
      noccoldb=nbe-ncorep
       if(core.ne.'corr  ') then
        nbasisold=nbasisold-ncore
        noccold=noccold-ncore
       endif
C Read the number of atomic domains
       if(localcc15p(localcc)) then
        ndom=nal-ncorep
       else
        open(unit=scrfile1,file='DOMAIN')
        rewind(scrfile1)
        read(scrfile1,*) ndom
        close(scrfile1)
       endif
       if (idomst-1.eq.ndom) then
         if (lccoporder.ne.'lccfirst') 
     $     write(iout,*) "Restarting local correlation computation",
     $  " results for all domains are read from file: localcc.restart"
         n1=1 ; n2=1 ; n3=1 ; label='ENERGY  '
         read(scrfile2,'(30a128)') (meth(imet),imet=1,nmet)
         go to 101 ! all LIS calculations are completed, jump to final result print
       endif
c
       if (lccrest.ne.'on      ') then 
         call lccorbnuminit(nocorr,nbcorr,nocorrmin,nocorrmax,nbcorrmin,
     $    nbcorrmax,noccold,nbasisold)
       elseif (lccrest.eq.'on      ') then
         write(iout,*) "Restarting local CC computation",
     $     " with LIS:", idomst
         write(iout,*) "   results from previous domains are read from",
     $     " file: localcc.restart"
       endif
c
       if (corembed.ne.'off     ') then
         call setenvcalc(envcalc,cscr16,minpfile,iout,lrc)
         hfenv=trim(envcalc).eq.'scf' 
         mp2env=trim(envcalc).eq.'mp2' 
         ccenv=lanycc(envcalc)
         if (ccenv) then 
          accprog=ccprog ! ccprog for the active orbitals
          eccprog=setccprog(envcalc) ! ccprog for the environment orbitals
          call getkey('laptol',6,alaptol,16) ! laptol for active orbitals
          elaptol='1.d-2           '  ! default laptol for the environment orbitals warning make it consistend with the default in minp.f
          laptoldefault=.false. ! is default set for laptol? 
          if (alaptol.eq.elaptol) laptoldefault=.true.       
         endif
         nact=0
       else
         mp2env=.false.
         hfenv=.false.
         ccenv=.false.
         lrc=.true.
         nact=noccolda
         laptoldefault=.true. ! not used
       endif
C Loop over domains
       do idom=idomst,ndom  
C Initialize files and variables
c {{{ localcc.eq.'2013'
         if(localcc.eq.'2013') then
          write(end,"(i9)") idom
          end=adjustl(end)
          beg='DOMAIN.' // end
          beg='mv ' // trim(beg) // ' DOMAIN'
          call ishell(beg)
          beg='MOCOEF.' // end
          beg='mv ' // trim(beg) // ' MOCOEF'
          call ishell(beg)
          beg='FOCK.' // end
          beg='mv ' // trim(beg) // ' FOCK'
          call ishell(beg)
          open(unit=scrfile1,file='DOMAIN',form='unformatted')
          rewind(scrfile1)
          read(scrfile1) nbasis,nocc
          read(scrfile1) 
          read(scrfile1) ncmo
c       write(6,*) 'nbasis,nocc,ncmo',nbasis,nocc,ncmo
C Calculate integrals
           if(trim(scfalg).ne.'disk'.or.ndom.gt.1.or..not.lseqc) then
            call runit('integ',intcpud,intwald)
            intcpu=intcpu+intcpud
            intwal=intwal+intwald
           endif
C Execute ovirt
          call runit('ovirt',ovirtcpud,ovirtwald)
          open(unit=ifcfile,file='iface',status='old')
          read(ifcfile,*) lmp2ened
          close(ifcfile)
          lmp2ene=lmp2ene+lmp2ened
          ovirtcpu=ovirtcpu+ovirtcpud
          ovirtwal=ovirtwal+ovirtwald
         else
           ncmo=1
         endif
c }}}
C Execute CC calculation for each occupied MO
         do iocc=1,ncmo
           ncoi=0 ! initilize domain as active
           if(localcc.eq.'2013') then
            read(scrfile1) (cmo,i=1,iocc)
            backspace(scrfile1)
           else
            cmo=idom
           endif

          csapprox=.false.
          write(ccmo,'(i6)') cmo
          ccmo=adjustl(ccmo)
          beg='55.' // trim(ccmo)
          beg=adjustl(beg)
          inquire(file=trim(beg),exist=log1)
          if (log1) then
           open(inp,status='old',file=trim(beg))
           read(inp,*) nbasis,nocc,idumy,idegen,idumy,ncoi 
           if (scftype.eq.'rohf ') then
            do i=1,6+nbasis**2
             read(inp,*)
            enddo
            read(inp,*) nob
            if (nob.eq.-1) then
             ! Closed-shell approximation in this domain
             nob=nocc/2
             csapprox=.true.
            endif
           else !scftype.ne.rohf
            nob=nocc/2
           endif
           rewind(inp) 
           nvirt=2*nbasis-nocc
          endif
          if (corembed.eq.'off     ') ncoi=0 ! initilize domain as active
c
          cycle_this_domain=(ncoi.eq.1.and.(mp2env.or.hfenv)) ! MP2/HF environment OR 0 virtual orbitals OR symmetry equivalent
     $                       .or.nvirt.eq.0.or.idegen.eq.0
c
c  prepartion of CC calculation in the LIS
          call preliscc(1,corembed,ncoi,inp,idegen,nact,ccenv,
     $ccprog,accprog,eccprog,nbasis,calctype,envcalc,log1,
     $cycle_this_domain,laptoldefault,elaptol,ifcfile,localcc,-1)
c
          if(log1.and.cycle_this_domain)  then
           write(iout,*)
           write(iout,122)
           write(iout,122)
           write(iout,"(' CC is not calculated for occupied orbital  ',
     $       i7,' ...')"),cmo
           write(iout,*)
           if (idegen.eq.0) then
             write(iout,*)'Symmetry equivalent LIS is already completed'
           else
             write(iout,*) 'Inactive LIS: 0 correlation E contribution'
           endif
           write(iout,122)
           cycle ! CC computation is not performed in the case of HF/MP2 environment model
          endif
c
c perform CC calculation in the LIS
           if(log1) then
             if(ccprog.eq.'ccsd') then
              call locfilecp(ccmo,'55       ')
               if(localcc.eq.'2013') then
                 call locfilemv(ccmo,'56      ')
                 call locfilemv(ccmo,'abcd.a  ')
                 call locfilemv(ccmo,'abcd.s  ')
                 call locfilemv(ccmo,'abci    ')
                 call locfilemv(ccmo,'abij    ')
                 call locfilemv(ccmo,'aibj    ')
                 call locfilemv(ccmo,'aijk    ')
                 call locfilemv(ccmo,'ijkl    ')
               else
                 call locfilecopy(0,cmo,naf,talg,nob,csapprox,0)
               endif
             endif ! ccprog.eq.'ccsd'
            end='cp ' // trim(beg) // ' fort.55' 
            call ishell(end)
            open(inp,status='old',file='fort.55')
            read(inp,*) nbasis,nocc,i,idegen
            close(inp)
            nbasis=2*nbasis
            nvirt=nbasis-nocc
C
            write(iout,*)
            write(iout,122)
            write(iout,122)
         write(iout,"(' Executing CC calculations for occupied orbital',
     $i4,' ...')"),cmo
            write(iout,*)
            write(iout,"(' Number of correlated electrons:      ',i4,3x,
     $'(',f5.1,'%)')"),nocc/2+nob,100.d0*dble(nocc/2+nob)/
     $dble(nal+nbe-2*ncorep)
            write(iout,"(' Number of correlated basis functions:',i4,3x,
     $'(',f5.1,'%)')"),nbasis/2,100.d0*dble(nbasis/2)/dble(nbasisold)
            write(iout,122)
c
c start the CC calculation in the LIS
            call ddmrcc(ccprog,csapprox)
c
c   process the iface file, restore keywords, save to localcc.restart
            if(trim(scftype).ne.'rhf') nocc=nocc/2+nob
            call postliscc(1,nocc,nbasis,meth,label,n1,n2,n3,eref,naftyp
     $,idegen,idomst,corembed,ncoi,ifcfile,
     $laptoldefault,alaptol,localcc,ccprog,maxmet, ! for localcc.restart
     $scrfile2,calctype,methact,
     $nmetact,nbasiss,noccs,ncore,idom,nmet,
     $nocorr,nbcorr,nocorrmin,nocorrmax,nbcorrmin,nbcorrmax,nact,
     $ecl,totcpu,totwal,lmp2ene,mp2corr,pairpllno,eenv,
     $edenergy(1),edenergy(2),edenergy(3),edenergy(4),edenergy(5),ccmem,
     $intcpu,intwal,scfcpu,scfwal,orbcpu,orbwal,ovirtcpu,ovirtwal,.true.
     $,-1)
c
c cleanup files of LIS idom 
            if (localcc15p(localcc).and.ccprog.eq.'ccsd') 
     $           call locfiledel(cmo,naf,talg)
           endif !log1
         enddo ! idom
        if(localcc.eq.'2013') close(scrfile1) ! close DOMAIN file
       enddo !idom
c
       close(scrfile2) ! close localcc.restart file 
c reset variables to their defaults in case of local embedding
       if (corembed.ne.'off     '.and.ccenv) then
           ccprog=accprog
           if (ccprog.eq.'mrcc') 
     $       call write56(nbasis,'none    ',calctype)
           call changekey('calc',4,calctype,16)
           nmet=nmetact
           meth(1:nmet)=methact(1:nmet)
           nact=noccolda
       endif
c cleanup temporary files of localcc
       end='rm -f DFINT_AI DFINT_AB DFINT_IJ 55'
       if (naf.ne.'off             ') end=trim(end) // ' ajb'
       if (talg.eq.'lapl'.or.talg.eq.'topr'.or.talg.eq.'to  '
     $     .or.talg.eq.'lato') end=trim(end) // ' laplbas'
       call ishell(end)
       if (ccprog.eq.'ccsd'.and.ccsdalg.eq.'disk    ') then
         end='rm -f abcd.a abcd.s abci abij aibj ijkl aijk 56'
         call ishell(end)
       endif
 101  continue
C Final results
      write(iout,*)
      write(iout,*)
      write(iout,122)
      write(iout,122)
      write(iout,*)
      write(iout,*) 'Final local CC results:'
      write(iout,*)
      
      if (nact.lt.1) nact=1
      write(iout,"(' Minimum number of correlated electrons:    ',
     $i6,7x,'(',f5.1,'%)')") nocorrmin,
     $100.d0*dble(nocorrmin)/dble(noccolda+noccoldb)
      write(iout,"(' Maximum number of correlated electrons:    ',
     $i6,7x,'(',f5.1,'%)')") nocorrmax,
     $100.d0*dble(nocorrmax)/dble(noccolda+noccoldb)
      write(iout,"(' Average number of correlated electrons:    ',
     $f8.1,5x,'(',f5.1,'%)')") dble(nocorr)/dble(nact),
     $100.d0*dble(nocorr)/dble(noccolda+noccoldb)/dble(nact)
      write(iout,"(' Minimum number of correlated orbitals:     ',
     $i6,7x,'(',f5.1,'%)')") nbcorrmin,
     $100.d0*dble(nbcorrmin)/dble(nbasisold)
      write(iout,"(' Maximum number of correlated orbitals:     ',
     $i6,7x,'(',f5.1,'%)')") nbcorrmax,
     $100.d0*dble(nbcorrmax)/dble(nbasisold)
      write(iout,"(' Average number of correlated orbitals:     ',
     $f8.1,5x,'(',f5.1,'%)')") dble(nbcorr)/dble(nact),
     $100.d0*dble(nbcorr)/dble(nbasisold)/dble(nact)
c
       do imet=1,nmet
        if(trim(meth(imet)).eq.'MP2') then
          mp2cpu=totcpu(imet)
          mp2wal=totwal(imet)
          imp2=imet
        elseif(trim(meth(imet)).eq.'CCSD') then
          totcpu(imet)=totcpu(imet)+mp2cpu ! MP2 part of the LIS CCSD calculation is added to CCSD
          totwal(imet)=totwal(imet)+mp2wal
          cccpu=totcpu(imet)
          ccwal=totwal(imet)
          iccsd=imet  
          if (lccoporder.eq.'lccfirst') then
            ovirtcpu=ovirtcpu-cccpu        ! time of CCSD was part of time of drpa (ovirt...) 
            ovirtwal=ovirtwal-ccwal
          endif
        elseif(trim(meth(imet)).eq.'CCSD(T)') then
          tcpu=totcpu(imet)                ! time of (T) was part of time of drpa (ovirt...) 
          twal=totwal(imet)
          ipart=imet
          if (lccoporder.eq.'lccfirst') then
            ovirtcpu=ovirtcpu-tcpu
            ovirtwal=ovirtwal-twal
          endif
        endif
      enddo
c
      write(iout,*)
      write(iout,"(' CPU time for integral calculations [min]:        ',
     $3x,f12.3)") intcpu/60.d0
      write(iout,"(' Wall time for integral calculations [min]:       ',
     $3x,f12.3)") intwal/60.d0
      write(iout,*)
      write(iout,"(' CPU time for SCF calculation [min]:              ',
     $3x,f12.3)") scfcpu/60.d0
      write(iout,"(' Wall time for SCF calculation [min]:             ',
     $3x,f12.3)") scfwal/60.d0
      write(iout,*)
      write(iout,"(' CPU time for orbital localization [min]:         ',
     $3x,f12.3)") orbcpu/60.d0
      write(iout,"(' Wall time for orbital localization [min]:        ',
     $3x,f12.3)") orbwal/60.d0
       if(localcc.eq.'2013') then
      write(iout,*)
      write(iout,"(' CPU time for domain construction [min]:          ',
     $3x,f12.3)") mulcpu/60.d0
      write(iout,"(' Wall time for domain construction [min]:         ',
     $3x,f12.3)") mulwal/60.d0
       endif
      write(iout,*)
      write(iout,"(' CPU time for integral transformations [min]:     ',
     $3x,f12.3)") ovirtcpu/60.d0
      write(iout,"(' Wall time for integral transformations [min]:    ',
     $3x,f12.3)") ovirtwal/60.d0
      write(iout,*)
      beg=' Reference energy [au]:'
      write(iout,7597) beg,escf
       if(lmp2dens.eq.'on ') then
        write(iout,*)
        beg=' L' // trim(ccorr) // ' correlation energy [au]:'
        write(iout,7597) beg,lmp2ene
        beg=' Total L' // trim(ccorr) // ' energy [au]:'
        write(iout,7597) beg,lmp2ene+escf
        write(iout,*)
         if(localcc.eq.'2013') then
           do imet=1,nmet
            if(trim(meth(imet)).eq.trim(ccorr))mp2corr=lmp2ene-ecl(imet)
           enddo
         endif
         if (naftyp.eq.'jpqpilot'.or.naftyp.eq.'jpq     ') then 
           beg=' Total MP2 correction for dropped NAFs [au]:'
           write(iout,7597) beg,mp2corr-pairpllno
         endif
        beg=' Total ' // trim(ccorr) // ' correction [au]:'
        write(iout,7597) beg,mp2corr
        if (corembed.ne.'off     ')
     $    call setenvcalc(envcalc,cscr16,minpfile,iout,lrc)
        if (trim(envcalc).eq.'mp2'.or.lanycc(envcalc)) then ! mp2env.or.ccenv
          call uppercase(envcalc,uenvcalc,16)
          beg=' Local ' // trim(uenvcalc) // 
     $                ' contribution of the environment [au]:'
          write(iout,7598) beg,eenv
          beg=' Total correction: ' // trim(ccorr) // 
     %        ' + environment [au]:'
          write(iout,7597) beg,eenv+mp2corr
          totcorr='all '
        else
          totcorr=ccorr
        endif
       endif
      write(iout,*)
C
      open(unit=ifcfile,file='iface',status='replace')
      write(ifcfile,"(68a)")
     $'#property method        sym st mul        value' //
     $'             CPU(sec)       Wall(sec)'
       do imet=1,nmet
         if(lmp2dens.eq.'off'.or.(trim(meth(imet)).ne.'MP2'.and.
     $trim(meth(imet)).ne.trim(ccorr))) then
           if(trim(meth(imet)).eq.'CCSD(T)') then
            beg=' CPU time for (T) corrections [min]:'
           else
            beg=' CPU time for ' // meth(imet)
            end=' calculations [min]:'
            call trimloc(beg,end)
           endif
          write(iout,"(a54,f11.3)") beg,totcpu(imet)/60.d0
          totcpu(imet)=totcpu(imet)+intcpu+scfcpu+orbcpu+mulcpu+ovirtcpu
           if(trim(meth(imet)).eq.'CCSD(T)') 
     $totcpu(imet)=totcpu(imet)+cccpu
          beg=' Total CPU time for ' // meth(imet)
          end=' [min]:'
          call trimloc(beg,end)
          write(iout,"(a54,f11.3)") beg,totcpu(imet)/60.d0
           if(trim(meth(imet)).eq.'CCSD(T)') then
            beg=' Wall time for (T) corrections [min]:'
           else
            beg=' Wall time for ' // meth(imet)
            end=' calculations [min]:'
            call trimloc(beg,end)
           endif
          write(iout,"(a54,f11.3)") beg,totwal(imet)/60.d0
          totwal(imet)=totwal(imet)+intwal+scfwal+orbwal+mulwal+ovirtwal
           if(trim(meth(imet)).eq.'CCSD(T)') 
     $totwal(imet)=totwal(imet)+ccwal
          beg=' Total wall time for ' // meth(imet)
          end=' [min]:'
          call trimloc(beg,end)
          write(iout,"(a54,f11.3)") beg,totwal(imet)/60.d0
          write(iout,*)
          if (trim(meth(imet)).eq.'CCSD(T)') then
           beg=' Total (T) correlation energy contribution [au]:' 
           write(iout,7597) beg,ecl(imet)-ecl(iccsd)
           if (localccXp(localcc,2018).and.corembed.eq.'off     ') then
          beg=' Total (T) correlation energy+0.5 MP2 correction [au]:'
            write(iout,7597) beg,ecl(imet)-ecl(iccsd)+mp2corr/2.d0
           endif
          elseif (trim(meth(imet)).eq.'CCSD'.and.localccXp(localcc,2018)
     $         .and.trim(calctype).eq.'ccsd(t)') then
           mp2fact=0.5d0
           mp2fact_halved=.true.
           if (totcorr.eq.ccorr) half='0.5'
          endif
          beg=' ' // meth(imet)
          end=' correlation energy [au]:'
          call trimloc(beg,end)
          write(iout,7597) beg,ecl(imet)
          beg=' Total ' // meth(imet)
          end=' energy [au]:'
          call trimloc(beg,end)
          write(iout,7597) beg,ecl(imet)+escf
           if(lmp2dens.eq.'on ') then
            beg=' ' // meth(imet)
            end=' correlation energy + '// 
     $trim(adjustl(half // ' ' // totcorr))// ' corrections [au]:'
            call trimloc(beg,end)
            write(iout,7597) beg,ecl(imet)+mp2fact*mp2corr+eenv
            beg=' Total LNO-' // meth(imet)
            end=' energy with '//trim(totcorr)//' corrections [au]:'
            call trimloc(beg,end)
            write(iout,7597) beg,ecl(imet)+escf+mp2fact*mp2corr+eenv
           endif
          write(iout,*)
          flush(iout) 
         endif
C Write iface file
        write(ifcfile,7596) label,meth(imet),n1,n2,n3,ecl(imet)+escf,
     $totcpu(imet),totwal(imet)
         if(lmp2dens.eq.'on ')
     $write(ifcfile,7596) label,
     $adjustl(trim(meth(imet)) // '+' // trim(ccorr)),n1,n2,n3,
     $ecl(imet)+escf+mp2fact*mp2corr+eenv,totcpu(imet),totwal(imet)
         if (mp2fact_halved) then 
          mp2fact=1.0d0
          mp2fact_halved=.false.
          if (totcorr.eq.ccorr) half='   '
         endif
       enddo
      write(iout,122)
      write(iout,122)
      close(ifcfile)
c
122   format(1x,70('='))
7596  format(a8,1x,a15,1x,3(i2,1x),1pe23.15,2(1pe15.8))
7597  format(a54,f20.12,a21)
7598  format(a57,f17.12)
C
      return
      end
C
************************************************************************
      subroutine dftd3(edisp,subsys) !HB
************************************************************************
* Interface to Grimme's dftd3 program to calculate dispersion correction
************************************************************************
#include "MRCCCOMMON"
      real*8 cput,walt
      character*2 subsys !HB
      character*6 cscr6
      character*8 embed !HB
      character*12 cscr12 !HB
      character*16 ccalc
      character*32 dft,cg
      character*256 edisp
C Shall we calculate gradient?
       if(dens.ge.2) then
        cg=' -grad | tee dftd3junk          '
       else
        cg=' | tee dftd3junk                '
       endif
C Call DFT-D3
       if(trim(edisp).eq.'auto') then
        call getkey('dft',3,dft,32)
        call getkey('calc',4,ccalc,16)
         if(trim(dft).eq.'b3lyp'.or.trim(dft).eq.'b3lyp3') then
          call runit('dftd3 COORD.xyz -func b3-lyp -bj'//cg,cput,walt)
          call edispprint('B3LYP-D3        ',cput,walt,subsys)
         else if(trim(dft).eq.'b3pw91') then
          call runit('dftd3 COORD.xyz -func b3pw91 -bj'//cg,cput,walt)
          call edispprint('B3PW91-D3       ',cput,walt,subsys)
         else if(trim(dft).eq.'bhlyp') then
          call runit('dftd3 COORD.xyz -func bh-lyp -bj'//cg,cput,walt)
          call edispprint('BHLYP-D3        ',cput,walt,subsys)
         else if(trim(dft).eq.'blyp') then
          call runit('dftd3 COORD.xyz -func b-lyp -bj'//cg,cput,walt)
          call edispprint('BLYP-D3         ',cput,walt,subsys)
         else if(trim(dft).eq.'bp86') then
          call runit('dftd3 COORD.xyz -func b-p -bj'//cg,cput,walt)
          call edispprint('BP86-D3         ',cput,walt,subsys)
         else if(trim(dft).eq.'pbe') then
          call runit('dftd3 COORD.xyz -func pbe -bj'//cg,cput,walt)
          call edispprint('PBE-D3          ',cput,walt,subsys)
         else if(trim(dft).eq.'pbe0') then
          call runit('dftd3 COORD.xyz -func pbe0 -bj'//cg,cput,walt)
          call edispprint('PBE0-D3         ',cput,walt,subsys)
         else if(trim(dft).eq.'hcth120') then
          call runit('dftd3 COORD.xyz -func hcth120 -bj'//cg,cput,walt)
          call edispprint('HCTH120-D3      ',cput,walt,subsys)
         else if(trim(dft).eq.'hyb_mgga_xc_mpw1b95') then
          call runit('dftd3 COORD.xyz -func mpw1b95 -bj'//cg,cput,walt)
          call edispprint('mPW1B95-D3      ',cput,walt,subsys)
         else if(trim(dft).eq.'tpss') then
          call runit('dftd3 COORD.xyz -func tpss -bj'//cg,cput,walt)
          call edispprint('TPSS-D3         ',cput,walt,subsys)
         else if(trim(dft).eq.'hyb_mgga_xc_tpssh') then
          call runit('dftd3 COORD.xyz -func tpssh -bj'//cg,cput,walt)
          call edispprint('TPSSh-D3        ',cput,walt,subsys)
         else if(trim(dft).eq.'b2plyp') then
          call runit('dftd3 COORD.xyz -func b2-plyp -bj'//cg,cput,walt)
          call edispprint('B2PLYP-D3       ',cput,walt,subsys)
         else if(trim(dft).eq.'b2gpplyp') then
         call runit('dftd3 COORD.xyz -func b2gp-plyp -bj'//cg,cput,walt)
          call edispprint('B2GPPLYP-D3     ',cput,walt,subsys)
         else if(trim(dft).eq.'dsdpbep86') then
          call ishell('echo "0.418 0.0000 0.000 5.65 0.0000 4" >' //
     $'$HOME/.dftd3par.`hostname`')
          call runit('dftd3 COORD.xyz -bj'//cg,cput,walt)
          call ishell('rm -f $HOME/.dftd3par.`hostname`')
          call edispprint('DSDPBEP86-D3    ',cput,walt,subsys)
         else if(trim(dft).eq.'dsdpbehb95') then
          call ishell('echo "0.58 0.0000 0.000 6.2 0.0000 4" >' //
     $'$HOME/.dftd3par.`hostname`')
          call runit('dftd3 COORD.xyz -bj'//cg,cput,walt)
          call ishell('rm -f $HOME/.dftd3par.`hostname`')
          call edispprint('DSDPBEhB95-D3   ',cput,walt,subsys)
         else if(trim(dft).eq.'drpa75') then
          call ishell('echo "0.375 0.0000 0.000 4.505 0.0000 4" >' //
     $'$HOME/.dftd3par.`hostname`')
          call runit('dftd3 COORD.xyz -bj'//cg,cput,walt)
          call ishell('rm -f $HOME/.dftd3par.`hostname`')
          call edispprint('dRPA75-D3       ',cput,walt,subsys)
         else if(trim(dft).eq.'off'.and.
     $         ccalc.eq.'scf             ') then
          call runit('dftd3 COORD.xyz -func hf -bj'//cg,cput,walt)
          call edispprint('HF-D3           ',cput,walt,subsys)
         else
          write(iout,*)
          write(iout,*) 'Empirical dispersion correction is not ' //
     $'implemented for this method!'
          call dmrccend(1)
         endif
       else
        open(minpfile,file='MINP')
!!! !HB
         if(subsys.eq.'no'.or.subsys.eq.'a1') then
          call getkeym('edisp',5,edisp,256)
          backspace(minpfile)
          read(minpfile,"(a6,a256)") cscr6,edisp
         else
          call getkeym('edisp_embed',11,edisp,256)
          backspace(minpfile)
          read(minpfile,"(a12,a256)") cscr12,edisp
         endif
!!!
        close(minpfile)
        edisp='dftd3 COORD.xyz ' // trim(edisp) // cg
        call runit(trim(edisp),cput,walt)
        call edispprint('Corrected       ',cput,walt,subsys)
       endif
C
      return
      end
C
************************************************************************
      subroutine edispprint(meth,cput,walt,subsys) !HB
************************************************************************
* Read and print dispersion correction
************************************************************************
#include "MRCCCOMMON"
      integer i,j
      real*8 edisp,temp,cput,walt,tmp
      character*2 subsys !HB
      character*6 cscr6
      character*8 cscr8
      character*15 cscr15
      character*16 meth,c16
      character*33 cscr32
C
      if(subsys.ne.'no') then !HB
       write(iout,*)
       if(subsys.eq.'ab') then
        write(iout,'(a)') ' D3 correction of the full system using the
     $ low-level method:'
       elseif(subsys.eq.'a2') then
        write(iout,'(a)') ' D3 correction of the embedded subsystem 
     $using the low-level method:'
       elseif(subsys.eq.'a1') then
        write(iout,'(a)') ' D3 correction of the embedded subsystem 
     $using the high-level method:'
       endif
      endif !HB
      write(iout,*)
      write(iout,"(1x,70('*'))")
      open(scrfile1,file='dftd3junk',status='old')
      cscr6='      '
       do while(cscr6.ne.'Edisp ')
        read(scrfile1,*,end=9876) cscr6
        cscr6=adjustl(cscr6)
       enddo
      backspace(scrfile1)
      read(scrfile1,"(a16,f11.4,f12.8)",end=9876) c16,tmp,edisp
      close(scrfile1,status='delete')
      write(iout,"(' Dispersion correction [au]:     ',f22.12)") edisp
      open(unit=ifcfile,file='iface',status='old')
      i=0
       do
        read(ifcfile,*,end=7597)
        i=i+1
       enddo
7597  rewind(ifcfile)
       do j=1,i-1
        read(ifcfile,*)
       enddo
      read(ifcfile,7596) cscr8,cscr15,i,i,i,ecc,temp,temp
      if(subsys.ne.'a2') then !HB
       ecc=ecc+edisp
      else
       ecc=ecc-edisp
      endif !HB
      cscr32=adjustl(trim(meth) // ' energy [au]:')
      if(subsys.ne.'no') then !HB
       write(iout,*)
       if(subsys.eq.'a1'.or.subsys.eq.'a2')
     $  cscr32='Final energy [au]:'
      endif !HB
      write(iout,"(1x,a32,f22.12)") cscr32,ecc
      cscr15=trim(cscr15) // '-D3'
      write(ifcfile,7596) cscr8,cscr15,i,i,i,ecc,cput,walt
      close(ifcfile)
7596  format(a8,1x,a15,1x,3(i2,1x),1pe23.15,2(1pe15.8))
C
      return
9876  write(iout,*) 
     $'Fatal error at the calculation of dispersion correcction!'
      call dmrccend(1)
      end
C
************************************************************************
      subroutine spoint(llg)
************************************************************************
* Driver for single-point calculations
************************************************************************
#include "MRCCCOMMON"
      integer i,j,m,nnb,itol,imult,nstate,istate,nstate2
      real*8 intcpu,intwal,orbcpu,orbwal,cput,walt,ovirtcpud,ovirtwald
      real*8 intcpud,intwald,scfcpu,scfwal
      real*8 lmp2ened,temp,edftbcor
      logical llg,log1,log2,log3
      character*4 localcc,cmult,ccprog,ctol,cscr4,molden,scfdtol,scftol
      character*4 ccsdrest
      character*4 schroute !HB
      character*5 scftype
      character*8 lccrest
      character*7 scfiguess,mcscfiguess
      character*8 cscr,scfalg,qscf
      character*8 iface,cscr8,qmmm,qmreg,dfintran,cialg,embed,grdens,rsh
      character*8 dual,lccoporder
      character*15 cscr15
      character*16 orblocc,orbloco,orblocv,calctype,dft,c16,orbloce!HB
      character*16 moselectalg !HB
      character*20 dfbasis_scf,dfbasis_cor
      character*256 edisp,edisp_embed !HB
      logical lseqc,lrpa,lrpanocc,localcc15p
C Initialize variables
      schroute='norm' !HB
      moselectalg='                '
      tim=0.d0
      intcpu=0.d0
      intwal=0.d0
      orbcpu=0.d0
      orbwal=0.d0
      ovirtcpud=0.d0
      ovirtwald=0.d0
      scfcpu=0.d0 
      scfwal=0.d0 
      lmp2ened=0.d0
      localcc='off '
      locno=0
      call ishell('rm -f ROUTE')
       if(llg) then
C Analyze the input file
        call getkey('iface',5,iface,8)
        call ishell('rm -f DOMAIN')
        call getkey('localcc',7,localcc,4)
        call getkey('lccrest',7,lccrest,8)
        call getkey('lccoporder',10,lccoporder,8)
        call getkey('ccprog',6,ccprog,4)  ! integ might change ccprog later
        call getkey('ccsdrest',8,ccsdrest,4)
        call getkey('calc',4,calctype,16)
        call getkey('dfbasis_scf',11,dfbasis_scf,20)
        call getkey('dfbasis_cor',11,dfbasis_cor,20)
        call getkey('qmmm',4,qmmm,8)
        call getkey('dfintran',8,dfintran,8)
        call getkey('nstate',6,cscr4,4)
        call getkey('grdens',6,grdens,8)
        call getkey('scfdtol',7,scfdtol,4)
        call getkey('scftol',6,scftol,4)
        call getkey('dual',4,dual,8)
        call getkey('scftype',7,scftype,5)
        call getkey('qscf',4,qscf,8)
        call getkey('mult',4,cmult,4)
        read(cscr4,*) nstate
        call getkey('nsing' ,5,cscr4,4)
        read(cscr4,*) i
        call getkey('ntrip' ,5,cscr4,4)
        read(cscr4,*) j
        if(nstate.ne.0.and.i.ne.0.and.j.ne.0) nstate2=i+j
        if(nstate.eq.0.and.i.ne.0.and.j.ne.0) nstate2=i+j
        if(nstate.ne.0.and.i.ne.0.and.j.eq.0) nstate2=i
        if(nstate.ne.0.and.i.eq.0.and.j.ne.0) nstate2=j+1
        if(nstate.eq.0.and.i.eq.0.and.j.ne.0) nstate2=j+1
        if(nstate.eq.0.and.i.ne.0.and.j.eq.0) nstate2=i  
        if(nstate.ne.0.and.i.eq.0.and.j.eq.0) nstate2=nstate
        nstate=nstate2
        if (lccrest.eq.'on      ')  goto 317
        if(lccrest .eq. 'off     ' .and. ccsdrest .eq. 'ccsd') goto 317
c        if (lccrest.eq.'restart ')  goto 318
         if(iface.eq.'none    ') then
          call getkey('embed',5,embed,8)
          call getkey('edisp_embed',11,edisp_embed,256) !HB
          call getkey('molden',6,molden,4)
          call getkey('scfiguess',9,scfiguess,7)
          if(scfiguess.eq.'off    '.or.scfiguess.eq.'lowlqm ') 
     $        call getvar('eref      ',eref)
C Run SCF calculation in small basis
           if(scfiguess.eq.'min    '.or.scfiguess.eq.'small  '.or.
     $dual.eq.'on      ') then
            if (scfiguess.ne.'off    ') then
             open(scrfile1,file='ROUTE')
              if(scfiguess.eq.'min    ') then
               write(scrfile1,"(a3)") 'min'
               call changekey('scfdtol',7,'6   ',4)
               call changekey('scftol',6,'5   ',4)
              else
               write(scrfile1,"(a3)") 'sml'
              endif
             close(scrfile1)
             call changekey('dfbasis_scf',11,'auto                ',20)
             call changekey('dfbasis_cor',11,'none                ',20)
              if(scfiguess.ne.'restart') 
     $       call changekey('scfiguess',9,'sad    ',7)
             call changekey('calc',4,'scf             ',16)
             write(iout,*) 
             write(iout,*)'Calculating integrals for small-basis SCF...'
             call runit('integ',intcpu,intwal)
             call ishell('rm -f DOMAIN')
             write(iout,*) 
             write(iout,*) 'Running small-basis SCF...'
              if(dual.ne.'off     ') then
               open(scrfile1,file='ROUTE')
               write(scrfile1,"(a3)") 'sml'
               close(scrfile1)
              endif
             call runit('scf',cput,walt)
             call changekey('dfbasis_scf',11,dfbasis_scf,20)
             call changekey('dfbasis_cor',11,dfbasis_cor,20)
             call changekey('calc',4,calctype,16)
            endif
            call changekey('scfiguess',9,'restart',7)
            if (scfiguess.eq.'off    ') then
             call getkey('scfiguess',9,scfiguess,7)
             write(iout,*) 
             write(iout,*) 'Small-basis SCF results are taken from the'
             write(iout,*) '   VARS and SCFDENSITIES files'
             call scfiguessOFFtasks(scfcpu,scfwal,2)
             write(iout,*) 
             write(iout,*) ' beginning the large basis calculation'
            endif
             if(scfiguess.eq.'min    ') then
              call changekey('scfdtol',7,scfdtol,4)
              call changekey('scftol',6,scftol,4)
             endif
           else if(dual.eq.'e1      '.or.dual.eq.'e2      ') then
            open(scrfile1,file='ROUTE')
            write(scrfile1,"(a3)") 'sml'
            close(scrfile1)
            call changekey('dfbasis_cor',11,'none                ',20)
            call changekey('calc',4,'scf             ',16)
           endif
C Calculate integrals
           call runit('integ',intcpu,intwal)
           call getkey('ccprog',6,ccprog,4)  ! integ might change ccprog
           if(scfiguess.eq.'off    ') then 
             call scfiguessOFFtasks(scfcpu,scfwal,1)
             goto 8765
           endif
9998      continue !HB
          if(schroute.eq.'term') then
           write(6,'(a)') ' ============================================
     $=========================='
           write(6,*)
           write(6,'(a)') ' *** New inital MOs will be generated using 
     $the low-level DFT ***'
           write(6,*)
           write(6,'(a)') ' ============================================
     $=========================='
           schroute='rest'
           scfiguess='restart' 
           call changekey('scfiguess',9,scfiguess,7)
          endif !HB
C DFT calculation in the case of embedding
           if(embed.ne.'off') then
            call get_moselectalg(moselectalg,minpfile)
            if(moselectalg.eq.'ecore') edisp_embed='off'
           endif
           if(embed.eq.'project '.or.embed.eq.'huzinaga'.or.
     $        embed.eq.'fdm     ') then
             if (scfiguess.ne.'lowlqm ') then
               open(scrfile1,file='ROUTE')
               write(scrfile1,"(a3)") 'em1'
               close(scrfile1)
               call runit('scf',cput,walt) ! run embedding SCF for the total QM system
             else  ! read the solution of embedding SCF for the total QM system
              open(varsfile,file='VARS',form='unformatted',
     $position='append')
              write(varsfile) 'edftab    ',ifltln,eref
              close(varsfile)
              write(iout,*)
              write(iout,*)' Read the embedding SCF solution from file'
              call scfiguessOFFtasks(scfcpu,scfwal,2)
             endif
             if(embed.ne.'fdm     ') then
              call getkey('orblocc',7,orblocc,16)
              call getkey('orbloco',7,orbloco,16)
              call getkey('orbloce',7,orbloce,16)
              call changekey('orbloco',7,orbloce,16)
              call changekey('orblocc',7,orbloce,16)
              if(moselectalg.ne.'ecore') then
               call runit('orbloc',orbcpu,orbwal)
              endif
              call runit('qmmod',orbcpu,orbwal)
              call changekey('orblocc',7,orblocc,16)
              call changekey('orbloco',7,orbloco,16)
               if(dual.eq.'e1      '.or.dual.eq.'e2      ')
     $call ishell('cp MOCOEF MOCOEF_AB')
             endif
             if(dual.ne.'e2      ') then
              open(scrfile1,file='ROUTE')
              write(scrfile1,"(a3)") 'em2'
              close(scrfile1)
              call changekey('scfiguess',9,'restart',7)
              call runit('scf',cput,walt)
              open(scrfile1,file='ROUTE')
              write(scrfile1,"(a3)") 'em3'
              close(scrfile1)
             endif
           else if(embed.eq.'sch     '.and.scfiguess.ne.'mo     ') then !HB
            open(scrfile1,file='ROUTE')
            write(scrfile1,"(a3)") 'em1'
            close(scrfile1)
            open(scrfile1,file='SCHINFO',form='unformatted') !HB
             if(schroute.eq.'norm') then
              write(scrfile1)
             else
              read(scrfile1)
             endif
             write(scrfile1) schroute
            close(scrfile1) !HB
            call runit('scf',cput,walt)
            call getkey('orblocc',7,orblocc,16) !HB
            call getkey('orbloco',7,orbloco,16) !HB
            call getkey('orbloce',7,orbloce,16) !HB
            call changekey('orblocc',7,orbloce,16) !HB
            call changekey('orbloco',7,orbloce,16) !HB
            if(moselectalg.ne.'ecore') then
             call runit('orbloc',orbcpu,orbwal) 
            endif
            call changekey('orblocc',7,orblocc,16) !HB
            call changekey('orbloco',7,orbloco,16) !HB
            call runit('qmmod',orbcpu,orbwal)
            call getvar('nfroz     ',nfroz) !HB
            open(scrfile1,file='SCHINFO',form='unformatted') !HB
             write(scrfile1) nfroz !HB
             write(scrfile1) schroute !HB
            close(scrfile1) !HB
            call changekey('scfiguess',9,'mo     ',7)
            open(scrfile1,file='ROUTE')
            write(scrfile1,"(a3)") 'sch'
            close(scrfile1)
           else if(embed.eq.'sch     '.and.scfiguess.eq.'mo     ') then !HB
            schroute='opti' 
            if(qmmm.eq.'off') then 
             open(scrfile1,file='SCHINFO',form='unformatted') !
              read(scrfile1) nfroz 
              write(scrfile1) schroute 
             close(scrfile1) 
            else
             open(scrfile1,file='mrcc_job.dat',status='old',
     $position='append')
              backspace(scrfile1)
              read(scrfile1,'(i8)') nfroz
             close(scrfile1)
             open(scrfile1,file='SCHINFO',form='unformatted')
              write(scrfile1) nfroz
              write(scrfile1) schroute
             close(scrfile1)
            endif
            open(varsfile,file='VARS',form='unformatted',
     $position='append')
             write(varsfile) 'nfroz     ',iintln,nfroz
            close(varsfile)
            open(scrfile1,file='ROUTE')
             write(scrfile1,"(a3)") 'sch'
            close(scrfile1) !HB
           endif
C SCF calculations
           if(trim(scftype).eq.'mcscf') then
            call getkey('mcscfiguess',11,mcscfiguess,7)
             if(trim(mcscfiguess).eq.'hf') then
              call changekey('calc',4,'scf             ',16)
               if(trim(cmult).eq.'1') then
                call changekey('scftype',7,'rhf  ',5)
               else
                call changekey('scftype',7,'rohf ',5)
               endif
              call changekey('qscf',4,'off     ',8)
              call runit('scf',cput,walt)
              call changekey('calc',4,calctype,16)
              call changekey('scftype',7,'mcscf',5)
              call changekey('scfiguess',9,'mo     ',7)
              call changekey('qscf',4,qscf,8)
             else
              call changekey('scfiguess',9,mcscfiguess,7)
             endif
           endif
           if(dual.ne.'e2      ')
     $    call runit('scf',cput,walt)
          scfcpu=cput
          scfwal=walt
          if(embed.eq.'sch') then !HB
           open(scrfile1,file='SCHINFO',form='unformatted')
            rewind(scrfile1)
            read(scrfile1)
            read(scrfile1) schroute
           close(scrfile1)
           if(schroute.eq.'term') then
             goto 9998
           endif
          endif !HB
C Dual basis set embedding
           if(dual.eq.'e1      '.or.dual.eq.'e2      ') then
             if(dual.eq.'e1      ')
     $      call ishell('mv MOCOEF MOCOEF_A')
            call ishell('rm -f DOMAIN')
            call changekey('dfbasis_cor',11,dfbasis_cor,20)
            call changekey('calc',4,calctype,16)
            call changekey('scfiguess',9,'restart',7)
            call getvar('nfroz     ',nfroz)
            call getvar('ncore     ',ncore)
            call runit('integ',intcpu,intwal)
            open(varsfile,file='VARS',form='unformatted',
     $position='append')
            write(varsfile) 'nfroz     ',iintln,nfroz
            write(varsfile) 'ncorenew  ',iintln,ncore
            close(varsfile)
            open(scrfile1,file='ROUTE')
            write(scrfile1,"(a4)") 'em1b'
            close(scrfile1)
             if(dual.eq.'e2      ') then
              open(scrfile1,file='FOCK',form='unformatted')
              read(scrfile1)
              read(scrfile1) ncore
              close(scrfile1)
             endif
            call runit('scf',cput,walt)
             if(dual.eq.'e2      ') then
              open(scrfile1,file='FOCK',form='unformatted')
              read(scrfile1) 
              write(scrfile1) ncore
              close(scrfile1)
             endif
c           call getkey('orblocc',7,orblocc,16)
c           call getkey('orbloco',7,orbloco,16)
c            if(orblocc.eq.'off             ')
c    $call changekey('orblocc',7,'pm              ',16)
c            if(orbloco.eq.'off             ')
c    $call changekey('orbloco',7,'pm              ',16)
c           call runit('orbloc',orbcpu,orbwal)
c           call changekey('orblocc',7,orblocc,16)
c           call changekey('orbloco',7,orbloco,16)
c           call runit('qmmod',orbcpu,orbwal)
            call ishell('mv SCFDENSITIES_A SCFDENSITIES')
            open(scrfile1,file='ROUTE')
            write(scrfile1,"(a4)") 'em2b'
            close(scrfile1)
            call runit('scf',cput,walt)
            open(scrfile1,file='ROUTE')
             if(dual.eq.'e1      ') then
              write(scrfile1,"(a4)") 'em3b'
              call ishell('mv SCFDENSITIES_At SCFDENSITIES')
             else
              write(scrfile1,"(a3)") 'em3'
             endif
            close(scrfile1)
            call runit('scf',cput,walt)
           endif
C
           if(trim(grdens).ne.'off'.and.
     $trim(calctype).ne.'scf') call ishell('cp DENSITY DENSITY.SCF')
8765      continue
           if(embed.eq.'fdm     '.and.trim(calctype).ne.'scf') then
            call ishell('rm -f DOMAIN')
            call runit('integ',intcpud,intwald)
           endif
          call getvar('rsh       ',rsh)
C Integral transformation for the CPHF equations in the case of DF
          call getkey('dens',4,cscr4,4)
          read(cscr4,*) dens
           if(dens.ne.0.and.dfbasis_scf.ne.'none                '.and.
     $        calctype.ne.'scf             '.and.
     $        trim(ccprog).ne.'cis') then
            call ishell('mv DFINV DFINV_SCF')
            open(scrfile1,file='ROUTE')
            write(scrfile1,"(a3)") 'scf'
            close(scrfile1)
            call runit('drpa',ovirtcpud,ovirtwald)
            call ishell('mv DFINT_AB DFINT_AB_SCF')
            call ishell('mv DFINT_AI DFINT_AI_SCF')
            call ishell('mv DFINT_IJ DFINT_IJ_SCF')
           endif
C Orbital localization
          call getkey('orblocc',7,orblocc,16)
          call getkey('orbloco',7,orbloco,16)
          call getkey('orblocv',7,orblocv,16)
           if(orblocc.ne.'off             '.or.
     $        orbloco.ne.'off             '.or.
     $       (orblocv.ne.'off             '.and.
     $        orblocv.ne.'pao             ')) then
             if(molden.ne.'off ') call ishell('cp MOLDEN MOLDEN.CAN')
            call ishell('cp MOCOEF MOCOEF.CAN')
            call runit('orbloc',orbcpu,orbwal)
            call ishell('cp MOCOEF MOCOEF.LOC') !NP
C Construction of frozen localized MOs for QM/MM calculation
             if(qmmm.eq.'amber  ') then
              open(minpfile,file='MINP')
              call getkeym('qmreg',5,qmreg,8)
              close(minpfile)
               if(qmreg.ne.'0       '.and.
     $            qmreg.ne.'        ') call runit('qmmod',orbcpu,orbwal)
             endif
           endif
         endif
c 318   continue ! restart here if lccrest=restart/domain NOT YET OK integ?
        lrpanocc=calctype.eq.'mp2             '.or.        ! call ldrpa.f, no CC calculation
     $           calctype.eq.'drpa            '.or.
     $           calctype.eq.'sosex           '.or.
     $           calctype.eq.'rpax2           '.or.
     $           calctype.eq.'rpa             '
         lrpa=lrpanocc.or.localcc15p(localcc)
c {{{    save timings of integ, scf, orbloc for localcc.restart ! NP
         if(lrpa.and.nstate.lt.2.and.lccrest.eq.'off     ') then
         open(varsfile,file='VARS',form='unformatted',position='append')
         write(varsfile) 'savedtimes',ifltln*6,
     $                         intcpu,intwal,scfcpu,scfwal,orbcpu,orbwal
         close(varsfile)
         endif
c }}}
         if(calctype.eq.'scf             '.and.nstate.lt.2) goto 1006
C Write fort.56 for ovirt and mrcc
        open(inp,status='old',file='fort.55')
        rewind(inp)
         if(iface.eq.'none    ') then
c         read(inp,*,end=1002) nbasis,nocc,ncore
          call getvar('nbf       ',nbf)
          call getvar('nal       ',nal)
          call getvar('nbe       ',nbe)
          call getvar('ncore     ',ncore)
          nbasis=nbf(1)
          nocc=nal+nbe
         else
          read(inp,*,end=1002) nbasis,nocc
          ncore=0
          call getkey('mult',4,cmult,4)
           if(cmult.eq.'    ') then
             if(mod(nocc,2).eq.0) then
              cmult='1   '
             else
              cmult='2   '
             endif
            call changekey('mult',4,cmult,4)
           endif
          read(cmult,*) imult
          nbe=(nocc-imult+1)/2
          nal=nbe+imult-1
         endif
         if(ccprog.ne.'cis '.and.(.not.lrpa.or.
     $      (localcc15p(localcc)
     $       .and.ccprog.eq.'mrcc'.and..not.lrpanocc)))
     $       call write56(nbasis,iface,calctype)
        close(inp)
         if(trim(ccprog).eq.'cis'.and.localcc.eq.'off ') then
         open(varsfile,file='VARS',form='unformatted',position='append')
           write(varsfile) 'localex   ',iintln,0
          close(varsfile)
         endif
C Calculate short/long-range integrals
         if(rsh.ne.'        ') then
          open(scrfile1,file='ROUTE')
          write(scrfile1,"(a4)") 'rsht'
          close(scrfile1)
          call ishell('cp VARS VARS.old')
          call runit('integ',intcpud,intwald)
          call ishell('mv VARS.old VARS')
          open(scrfile1,file='ROUTE')
          write(scrfile1,"(a4)") 'rsht'
          close(scrfile1)
          call runit('drpa',ovirtcpud,ovirtwald)
          open(scrfile1,file='ROUTE')
c         write(scrfile1,"(a4)") 'rsht' !szemet
          write(scrfile1,"(a3)") 'rsh'
          close(scrfile1)
C DFINT_AI_RS  - RS (ai|P) alpha
C DFINT_AIb_RS - RS (ai|P) beta
C DFINT_PQ_RS  - RS (P|Q)
C DFINV        - Cholesky deco of normal (P|Q)
         endif
C
         if((localcc.eq.'off '.or.lrpa).and.iface.eq.'none    ') then
C Calculate DF integrals for correlation calculation
          call getkey('dfbasis_cor',11,dfbasis_cor,20)
          call getvar('lseqc     ',lseqc)
          call getkey('scfalg',6,scfalg,8)
          call getkey('cialg',5,cialg,8)
           if(((.not.lseqc.or.dens.ne.0.or.rsh.ne.'        ').and.
     $dfbasis_cor.ne.'none                ').or.
     $trim(scfalg).ne.'disk'.or.(lrpa.and.localcc.ne.'off ').or.
     $(dfbasis_scf.ne.'none                '.and.
     $ dfbasis_cor.eq.'none                ')) then
c            if (localcc.eq.'off '.or.
c    $       (localcc.eq.'2015'.or..not.localcc15p(localcc))) then ! skip integ if localcc>2015
              call ishell('cp VARS VARS.old')
              call runit('integ',intcpud,intwald)   
              intcpu=intcpu+intcpud                 
              intwal=intwal+intwald                 
c             call getvar('nbf       ',nbf)
              call ishell('mv VARS.old VARS')
c             open(varsfile,file='VARS',form='unformatted',
c    $position='append')
c             write(varsfile) 'nbfnew    ',iintln*nbset,nbf
c             close(varsfile)
c            endif ! localcc.eq.'off '.or.
             if(rsh.ne.'        ') then
              open(scrfile1,file='ROUTE')
              write(scrfile1,"(a3)") 'rsh'
              close(scrfile1)
             endif
           endif
          call ishell('rm -f DOMAIN')
C Solve CIS equations in the case of local excited-state calculation
           if(localcc.ne.'off '.and.nstate.gt.1) then
            call ishell('cp VARS VARS.old')
            open(varsfile,file='VARS',form='unformatted',
     $position='append')
C 0: canonical, 1: CIS for local, 2: CC for local
             write(varsfile) 'localex   ',iintln,1
            close(varsfile)
C
            call ishell('cp MOCOEF MOCOEF.LOC')
            call getkey('redcost_tddft',13,c16,16)
            if(trim(c16).eq.'off') call ishell('cp MOCOEF.CAN MOCOEF')
            if(trim(cialg).eq.'disk') 
     $call runit('drpa',ovirtcpud,ovirtwald)
            call runit('cis',cput,walt)
C
            call ishell('mv VARS.old VARS')
            open(varsfile,file='VARS',form='unformatted',
     $position='append')
C 0: canonical, 1: CIS for local, 2: CC for local
             write(varsfile) 'localex   ',iintln,2
            close(varsfile)
C
           endif
           if(lrpa) then
C DF-MP2, RPA, or local correlation (2015-19) calculation
             if(localcc.ne.'off '.and.nstate.gt.1) then
              do istate=2,nstate
               open(scrfile1,file='istate')
               write(scrfile1,*) istate,nstate
               close(scrfile1)
               call ishell('cp MOCOEF.LOC MOCOEF')
               call runit('drpa',ovirtcpud,ovirtwald)
C DAVID WARNING!!!
               call ishell('cp MOCOEF.CAN MOCOEF')
               call runit('cis',cput,walt)
              enddo
             else
              call runit('drpa',ovirtcpud,ovirtwald) ! ground state local correlation
             endif
             if(localcc.eq.' off'.or.localcc.eq.'2013'.or.
     $          (calctype.ne.'ccsd            '
     $.and.calctype.ne.'ccsd(t)         '.and.ccprog.eq.'ccsd').or.
     $ (ccprog.eq.'mrcc'.and.lrpanocc).or.
     $(localcc.ne.'off '.and.nstate.gt.1)) goto 1006
           else if(dfbasis_cor.ne.'none                '.and.
     $(ccprog.eq.'mrcc'.or.ccprog.eq.'cis '.or.
     $(ccprog.eq.'ccsd'.and.dfintran.eq.'drpa    '))) then
C DF integrals for mrcc or cis
             if(ccprog.ne.'cis '.or.cialg.eq.'disk   '.or.
     $((trim(cialg).eq.'direct'.or.trim(cialg).eq.'direct2'.or.
     $trim(cialg).eq.'direct3').and.(trim(calctype).ne.'cis'.and.
     $trim(calctype).ne.'scf'))) then
              call runit('drpa',cput,walt)
             endif
           else
C Integral transformation and orbital optimization
            call runit('ovirt',ovirtcpud,ovirtwald)
           endif
         endif
         if(localcc.ne.'off ') locno=3
       endif
C
      if(.not.localcc15p(localcc).and.trim(ccprog).ne.'cis') then
       open(inpfile,status='unknown',file='fort.56')
       read(inpfile,*,end=1002) op,nsing,ntrip,rest,calc,i,conver,i,i,
     $iclsh,i,ihf,ndoub,nacto,nactv,itol,maxex,sacc,ptfreq,dboc,maxcor
c    $,locno
       op1=op
       nroot=nsing+ntrip+ndoub
       close(inpfile)
      endif 
       if(calc.gt.10) then
        ssmrcc=calc-10
        calc=1
       else
        ssmrcc=0
       endif
C
      inquire(file='fort.55',exist=log1)
      inquire(file='MDCINT', exist=log2)
      inquire(file='MRCONEE',exist=log3)
       if(.not.log1) goto 1002
c      if(.not.log1.and.(.not.log2.or..not.log3)) goto 1002
c      if(log2.and.log3) then
c       open(mrconeefile,file='MRCONEE',form='UNFORMATTED')
c       read(mrconeefile) nbasis,log,nuc
c       call dirac2mrcc(nbasis)
c       close(mrconeefile)
c      endif
      open(inp,status='unknown',file='fort.55')
      rewind(inp)
      read(inp,*,end=1002) m
       if(m.ne.0) then
        rewind(inp)
       endif
      read(inp,*,end=1002) nbasis,nocc
      read(inp,*,end=1002) (m,i=1,nbasis)
      read(inp,*,end=1002) m
      close(inp)
       if(m.eq.-3.and.(.not.log2.or..not.log3)) goto 1002
      nnb=nbasis
      nbasis=2*nbasis
      nvirt=nbasis-nocc
C
C Local CC calculation
 317   continue 
       if(localcc.ne.'off ') then
        call locdrv(intcpu,intwal,orbcpu,orbwal,localcc,ovirtcpud,
     $ovirtwald,lmp2ened,scfcpu,scfwal,calctype) 
       else
C Integral transformation and orbital optimization
c        if(locno.gt.0) call runit('ovirt',ovirtcpud,ovirtwald)
C CC calculation
         if(ssmrcc.gt.0) then
          call dssmrcc(nnb,ccprog)
         else
          call ddmrcc(ccprog,.false.)
         endif
       endif
C
1006  continue
       if(llg) then
C Calculate empirical dispersion
        call getkey('edisp',5,edisp,256)
        if(embed.eq.'off') then !HB
         if(trim(edisp).ne.'off') call dftd3(edisp,"no")
        else
         if(trim(edisp_embed).ne.'off') then
          call dftd3(edisp_embed,"ab")
          if(dens.ge.2) call ishell('cp dftd3_gradient dftd3_grad_ab')
         endif
         call ishell('cp COORD.xyz COORD_save.xyz')
         call ishell('cp COORD_SUBSYSA.xyz COORD.xyz')
         if(trim(edisp_embed).ne.'off') then
          call dftd3(edisp_embed,"a2")
          if(dens.ge.2) call ishell('cp dftd3_gradient dftd3_grad_a2')
         endif
         if(trim(edisp).ne.'off'.and.moselectalg.ne.'ecore') then
          call dftd3(edisp,"a1")
          if(dens.ge.2) call ishell('cp dftd3_gradient dftd3_grad_a1')
         elseif(trim(edisp).ne.'off'.and.moselectalg.eq.'ecore') then
          call dftd3(edisp,"no")
         endif
         call ishell('cp COORD_save.xyz COORD.xyz')
        endif !HB
C Gradient and property calculation
         if(iabs(dens).ge.1.and.iabs(dens).le.2) then
           if(dfbasis_scf.ne.'none                '.and.
     $calctype.ne.'scf             '
     $.and.trim(ccprog).ne.'cis') call ishell('mv DFINV_SCF DFINV')
          call runit('prop',cput,walt)
         endif
C Write Qm energy for QM/MM calculations
         if(qmmm.eq.'charmm  ') then
C CHARMM
          call getenergy(temp,cscr15)
          open(scrfile1,file='q1.out')
          rewind(scrfile1)
          write(scrfile1,"(' 11 ',e23.15,' Convergence criterion met')")
     $temp
          close(scrfile1)
         else if(qmmm.eq.'amber  ') then
C Amber
C         Write energy of the QM region to the mrcc_job.dat
          call getenergy(temp,cscr15)
           if(dens.eq.2) then
            open(scrfile1,file='mrcc_job.dat',status='old',
     $position='append')
           else
            open(scrfile1,file='mrcc_job.dat',status='replace')
           endif
          write(scrfile1,'(a)')' Final energy of the QM region [AU]'
          write(scrfile1,'(e23.15)') temp
          close(scrfile1)
C         Write self energy of the point charges to the mrcc_job.dat
          call getvar('selfenergy',temp)
          open(scrfile1,file='mrcc_job.dat',status='old',
     $position='append')
          write(scrfile1,'(a)')' Self energy of the point charges [AU]'
          write(scrfile1,'(e23.15)') temp
          close(scrfile1)
         endif
       endif
c restore the canonical orbitals to the MOCOEF file !NP
      inquire(file='MOCOEF.LOC',exist=log1)
      inquire(file='MOCOEF.CAN',exist=log2)
       if(log1.and.log2) call ishell('cp MOCOEF.CAN MOCOEF')
C
      return
1002  continue
      write(iout,*) 'Input file does not exist!'
      call flush(iout)
      call dmrccend(1)
      end
C
************************************************************************
      subroutine scfiguessOFFtasks(scfcpu,scfwal,task)
************************************************************************
* perfrom necessary tasks is the scf calculation is skipped due to scfiguess=off keyword
************************************************************************
#include "MRCCCOMMON"
      character*16 dft
      real*8 scfcpu,scfwal
      integer task
C
      if (task.eq.1) then
        open(varsfile,file='VARS',form='unformatted',
     $  position='append')
        write(varsfile) 'eref      ',ifltln,eref
        close(varsfile)
        open(unit=ifcfile,file='iface',status='unknown')
        write(ifcfile,"(68a)")
     &  '#property method        sym st mul        value' //
     &  '             CPU(sec)       Wall(sec)'
        write(ifcfile,7596) 'ENERGY  ','SCF            ',1,1,1,eref,
     &  0.d0,0.d0
        close(ifcfile)
7596    format(a8,1x,a15,1x,3(i2,1x),1pe23.15,2(1pe15.8))
      endif
c
      if (task.eq.1.or.task.eq.2) then
       call getkey('dft',3,dft,16)
       write(*,*)
       if(dft.eq.'off             ') then
       WRITE(*,"(' ***FINAL HARTREE-FOCK ENERGY:',F24.16,' [AU]')") eref
       else
       WRITE(*,"(' ***FINAL KOHN-SHAM ENERGY:   ',F24.16,' [AU]')") eref
       endif
       scfcpu=0.d0
       scfwal=0.d0
      endif
c
      return
      end
C
************************************************************************
      subroutine composite(llg)
************************************************************************
* Performs calculations with composite methods, e.g., ONIOM
************************************************************************ 
      implicit none
      character*8 gopt,freq,oniom,qmmm
      logical llg
      call getkey('gopt',4,gopt,8)
      call getkey('freq',4,freq,8)
      call getkey('oniom',5,oniom,8)
      call getkey('qmmm',4,qmmm,8)
      if(oniom.ne.'off') then
       call oniomcore(oniom,gopt,freq,qmmm)
      endif
      return
      end
C
